open GccTypes

let block_label = "block_label"

(*

   A block is a closure, whose label is "block", and whose
   frame contains:
     position 0 -> TAG
     position 1 -> SIZE
     position 3+n -> var[n]

  The block is called with two arguments METH and IDX
  METH = 0 => get
  METH = 1 => set
  METH = 2 => size
  METH = 3 => tag

  The field access is log2(size).

*)

let new_block tag fields cont =
  let size = List.length fields in
  LDC tag ::
  LDC size ::
  List.fold_right (@@) fields
  (LDF (L "create_block") ::
   AP (2+size) ::
   cont)

let get_field block n cont =
  LDC n ::
  LDC 0 :: (* read *)
  block @ AP 2 :: cont

let set_field block n value cont =
  value @
  LDC n ::
  LDC 1 :: (* write *)
  block @ AP 2 :: cont

let get_dyn_field block field cont =
  field @
  LDC 0 :: (* read *)
  block @ AP 2 :: cont

let set_dyn_field block field value cont =
  value @ field @
  LDC 1 :: (* write *)
  block @ AP 2 :: cont

let get_size block cont =
  LDC 0 :: (* dummy *)
  LDC 2 :: (* size *)
  block @ AP 2 :: cont

let get_tag block cont =
  LDC 0 :: (* dummy *)
  LDC 3 :: (* tag *)
  block @ AP 2 :: cont

let block_name min max =
  if max > min then
    Printf.sprintf "block_%d_%d" min max
  else
    Printf.sprintf "block_%d" min

let block_access_name max =
  Printf.sprintf "block_%d_access" max

let rec block_access min max =
  if max = min then
    let bl = block_name min min in
    [
      LABEL (L bl);
      LD (0,1); (* operation *)
      LDC 0;
      CEQ;
      TSEL (L (bl ^ "_read"), L (bl ^ "_write"));

      LABEL (L (bl ^ "_read"));
      LD (1, min+2);
      RTN;

      LABEL (L (bl ^ "_write"));
      ST (1, min+2);
      LDC 0;
      RTN;
    ]
  else
    let mean = (min + max) / 2 in
    [
      LABEL (L (block_name min max));
      LD (0,0); (* field *)
      LDC (mean+1);
      CGTE;
      TSEL (L (block_name (mean+1) max), L(block_name min mean));
    ]
    @
    block_access min mean
    @
    block_access (mean+1) max

let block max =
  let access size =
    [ LABEL (L (block_access_name size));
      LD (0,1); (* operation *)
      LDC 1;
      CGT;
      TSEL (L ("block_tag_size"), L (block_name 0 size)) ] in
  let create_lbl size = L (Printf.sprintf "create_block_%d_max" size) in
  let create size =
    [ LABEL (create_lbl size); LDF (L (block_access_name size)); RTN ] in
  [ LABEL (L "create_block");
    LD (0, 1); (* size *)
    LDC 1;
    CEQ;
    TSEL (create_lbl 0, (L "create_block_1_min"));
    LABEL (L "create_block_1_min");
    LD (0, 1); (* size *)
    LDC 2;
    CEQ;
    TSEL (create_lbl 1, (L "create_block_2_min"));
    LABEL (L "create_block_2_min");
    LDC 4;
    LD (0, 1); (* size *)
    CGTE;
    TSEL (create_lbl 3, (L "create_block_4_min"));
    LABEL (L "create_block_4_min");
    LDC 8;
    LD (0, 1); (* size *)
    CGTE;
    TSEL (create_lbl 7, (L "create_block_8_min"));
    LABEL (L "create_block_8_min");
    LDC 16;
    LD (0, 1); (* size *)
    CGTE;
    TSEL (create_lbl 15, (L "create_block_16_min"));
    LABEL (L "create_block_16_min");
    LDC 32;
    LD (0, 1); (* size *)
    CGTE;
    TSEL (create_lbl 31, (L "create_block_32_min"));
    LABEL (L "create_block_32_min");
    LDC 64;
    LD (0, 1); (* size *)
    CGTE;
    TSEL (create_lbl 63, (L "create_block_64_min"));
    LABEL (L "create_block_64_min");
    LDC 128;
    LD (0, 1); (* size *)
    CGTE;
    TSEL (create_lbl 127, create_lbl 255);
 ]
  @ List.flatten (List.map create [0; 1; 3; 7; 15; 31; 63; 127; 255])
  @ List.flatten (List.map access [0; 1; 3; 7; 15; 31; 63; 127; 255])
  @ [
    LABEL (L ("block_tag_size"));
    LD (0,1); (* operation *)
    LDC 2;
    CEQ;
    TSEL (L ("block_size"), L ("block_tag"));

    LABEL (L ("block_tag"));
    LD (1, 0);
    RTN;

    LABEL (L ("block_size"));
    LD (1, 1);
    RTN;
  ]
  @ block_access 0 max

let array_name min max =
  if max > min then
    Printf.sprintf "array_%d_%d" min max
  else
    Printf.sprintf "array_%d" min

let rec array min max =
  (*  Printf.eprintf "block %d %d\n%!" min max; *)
  if min < max then begin
    let mean = (min + max) / 2 in
    if mean > min then
      [
        LABEL (L (array_name min max));
        LD (0,1); (* field *)
        LDC mean;
        CGTE;
        TSEL (L (array_name mean max), L(array_name min (mean-1)));
      ]
      @
      array min (mean-1)
      @
      array mean max
    else
      [
        LABEL (L (array_name min max));
        LD (0,1); (* field *)
        LDC min;
        CGT;
        TSEL (L (array_name (min+1) max), L (array_name min min));
      ]
      @
      array min min
      @
      array (min+1) max

  end else begin
    let bl = array_name min min in
    [
      LABEL (L bl);
      LD (0,0); (*bwaaa*)
      LDF (L "array_fun");
      AP (min+1);
    ]
    @
    [ LABEL (L (bl ^ "_read"));
      LD (1, min);
      RTN ]
    @
    [ LABEL (L (bl ^ "_write"));
      LD (0, 2);
      ST (1, min);
      RTN ]
  end

let compare_lbl = L "caml_compare"
let eq_lbl = L "caml_equal"
let neq_lbl = L "caml_notequal"
let gt_lbl = L "caml_greaterthan"
let ge_lbl = L "caml_greaterequal"
let lt_lbl = L "caml_lessthan"
let le_lbl = L "caml_lessequal"

let alloc s =
  let name = string_of_int s in
  [ LABEL (L ("caml_make_vect_" ^ name ^ "_create"));
    LDC s;
    LD (1, 4); (* SIZE *)
    SUB;
    ST (1, 0); (* i *)
    LABEL (L ("caml_make_vect_" ^ name ^ "_loop"));
    LDC 0;
    LD (1, 0); (* i *)
    CGTE;
    TSEL (L ("caml_make_vect_" ^ name ^ "_alloc"),
          L ("caml_make_vect_" ^ name ^ "_push"));
    LABEL (L ("caml_make_vect_" ^ name ^ "_push"));
    LDC 0;
    LD (1, 0); (* i *)
    LDC 1;
    SUB;
    ST (1, 0); (* i *)
    LDC 0;
    TSEL (L ("caml_make_vect_" ^ name ^ "_loop"),
          L ("caml_make_vect_" ^ name ^ "_loop"));
    LABEL (L ("caml_make_vect_" ^ name ^ "_alloc"));
    LDF (L "create_block");
    TAP (s+2);
  ]

let caml_make_vect =
  [ LABEL (L "caml_make_vect");
    LD (0, 4); (* size *)
    ST (0, 0); (* i *)
    LDC 0; (* tag *)
    LD (0, 4); (* size *)
    LABEL (L "caml_make_vect_loop");
    LDC 0;
    LD (0, 0); (* i *)
    CGTE;
    TSEL (L "caml_make_vect_alloc", L "caml_make_vect_push");
    LABEL (L "caml_make_vect_push");
    LD (0, 5); (* init *)
    LD (0, 0); (* i *)
    LDC 1;
    SUB;
    ST (0, 0); (* i *)
    LDC 0;
    TSEL (L "caml_make_vect_loop", L "caml_make_vect_loop");
    LABEL (L "caml_make_vect_alloc");
    LDF (L "caml_make_vect_create");
    TAP 0;
    LABEL (L "caml_make_vect_create");
    LD (1, 4); (* size *)
    LDC 128;
    CGT;
    TSEL (L "caml_make_vect_256_create", L "caml_make_vect_max_128_create");
    LABEL (L "caml_make_vect_max_128_create");
    LD (1, 4); (* size *)
    LDC 64;
    CGT;
    TSEL (L "caml_make_vect_128_create", L "caml_make_vect_max_64_create");
    LABEL (L "caml_make_vect_max_64_create");
    LD (1, 4); (* size *)
    LDC 32;
    CGT;
    TSEL (L "caml_make_vect_64_create", L "caml_make_vect_max_32_create");
    LABEL (L "caml_make_vect_max_32_create");
    LD (1, 4); (* size *)
    LDC 16;
    CGT;
    TSEL (L "caml_make_vect_32_create", L "caml_make_vect_max_16_create");
    LABEL (L "caml_make_vect_max_16_create");
    LD (1, 4); (* size *)
    LDC 8;
    CGT;
    TSEL (L "caml_make_vect_16_create", L "caml_make_vect_max_8_create");
    LABEL (L "caml_make_vect_max_8_create");
    LD (1, 4); (* size *)
    LDC 4;
    CGT;
    TSEL (L "caml_make_vect_8_create", L "caml_make_vect_4_create");
 ] @ List.flatten (List.map alloc [4;8;16;32;64;128;256])

let rec pow2 n = if n <= 0 then 1 else 2 * (pow2 (pred n))

let bit_extract n value =
  [ value;
    LDC (pow2 n);
    DIV;
    value;
    LDC (pow2 (n+1));
    DIV;
    LDC 2;
    MUL;
    SUB;
  ]

type bit_comp = BcAnd | BcOr | BcXor

let bit_compare = function
  | BcAnd -> [MUL]
  | BcOr -> [ADD; LDC 0; CGT;]
  | BcXor -> [ADD; LDC 1; CEQ;]

let bit_contract = [LDC 2; MUL; ADD]

let bit_compare_8values comp a b cont =
  (bit_extract 0 a) @ (bit_extract 0 b) @
  (bit_compare comp) @
  (bit_extract 1 a) @ (bit_extract 1 b) @
  (bit_compare comp) @
  (bit_extract 2 a) @ (bit_extract 2 b) @
  (bit_compare comp) @
  (bit_extract 3 a) @ (bit_extract 3 b) @
  (bit_compare comp) @
  (bit_extract 4 a) @ (bit_extract 4 b) @
  (bit_compare comp) @
  (bit_extract 5 a) @ (bit_extract 5 b) @
  (bit_compare comp) @
  (bit_extract 6 a) @ (bit_extract 6 b) @
  (bit_compare comp) @
  (bit_extract 7 a) @ (bit_extract 7 b) @
  (bit_compare comp) @
  bit_contract @ bit_contract @ bit_contract @ bit_contract @
  bit_contract @ bit_contract @ bit_contract (* @ bit_contract *) @ cont

let caml_compare =
  let ints_lbl = L "compare_ints" in
  let noints_lbl = L "compare_noints" in
  let blocks_lbl = L "compare_blocks" in
  let different_lbl = L "compare_different_int" in
  let compare_fields_lbl = L "compare_fields" in
  let compare_size_lbl = L "compare_size" in
  let compare_equal_lbl = L "compare_equal_lbl" in
  let compare_field_lbl = L "compare_field" in
  let continue_lbl = L "compare_continue" in
  let return_lbl = L "compare_return" in
  let loop_test_lbl = L "compare_loop_test" in
  [ LABEL compare_lbl;
    LD (0, 4); (* x *)
    ATOM;
    LD (0, 5); (* y *)
    ATOM;
    ADD;
    ST (0, 0); (* Atom(x) + Atom(y) *)
    LD (0, 0);(* Atom(x) + Atom(y) *)
    LDC 2;
    CEQ;
    TSEL (ints_lbl, noints_lbl);
    LABEL ints_lbl;
    LD (0, 4); (* x *)
    LD (0, 5); (* y *)
    SUB;
    RTN;
    LABEL noints_lbl;
    LD (0, 0); (* Atom(x) + Atom(y) *)
    LDC 0;
    CEQ;
    TSEL (blocks_lbl, different_lbl);
    LABEL blocks_lbl ] @
    get_size [LD (0, 4)] @@ (* x *)
  [ ST (0, 1) ] @           (* size (x) *)
    get_size [LD (0, 5)] @@ (* y *)
  [ ST (0, 2);     (* size (y) *)
    LD (0, 1);     (* size (x) *)
    LD (0, 2);     (* size (y) *)
    CEQ;
    TSEL (compare_fields_lbl, compare_size_lbl);
    LABEL compare_size_lbl;
    LD (0, 1);     (* size (x) *)
    LD (0, 2);     (* size (y) *)
    SUB;
    RTN;
    LABEL compare_fields_lbl;
    LDC 0;
    ST (0, 2); (* i *)
    LABEL loop_test_lbl;
    LD (0, 1); (* size *)
    LD (0, 2); (* i *)
    CEQ;
    TSEL (compare_equal_lbl, compare_field_lbl);
    LABEL compare_equal_lbl;
    LDC 0;
    RTN;
    LABEL compare_field_lbl;
    LDC 0; (* Start recursive call *)
    LDC 0;
    LDC 0;
    LDC 0;
    LD (0, 2); (* i *)
    LDC 0;     (* GET *)
    LD (0, 4); (* x *)
    AP 2;      (* -> x(i) *)
    LD (0, 2); (* i *)
    LDC 0;     (* GET *)
    LD (0, 5); (* y *)
    AP 2;      (* -> y(i) *)
    LDF compare_lbl;
    AP 6; (* rec call *)
    ST (0, 0); (* rec call x(i) y(i) *)
    LD (0, 0); (* rec call x(i) y(i) *)
    LDC 0;
    CEQ;
    TSEL (continue_lbl, return_lbl);
    LABEL return_lbl;
    LD (0, 0); (* rec call x(i) y(i) *)
    RTN;
    LABEL continue_lbl;
    LD (0, 2); (* i *)
    LDC 1;
    ADD;
    ST (0, 2); (* i *)
    LDC 0;
    TSEL (loop_test_lbl, loop_test_lbl);
    LABEL different_lbl;
    LD (0, 5); (* y *)
    ATOM;
    LD (0, 4); (* x *)
    ATOM;
    SUB;
    RTN;
    (* EQ *)
    LABEL eq_lbl;
    LDC 0;
    LDC 0;
    LDC 0;
    LDC 0;
    LD (0, 4);
    LD (0, 5);
    LDF compare_lbl;
    AP 6;
    LDC 0;
    CEQ;
    RTN;
    (* NEQ *)
    LABEL neq_lbl;
    LDC 1;
    LDC 0;
    LDC 0;
    LDC 0;
    LDC 0;
    LD (0, 4);
    LD (0, 5);
    LDF compare_lbl;
    AP 6;
    LDC 0;
    CEQ;
    SUB;
    RTN;
    (* lt *)
    LABEL lt_lbl;
    LDC 1;
    LDC 0;
    LDC 0;
    LDC 0;
    LDC 0;
    LD (0, 4);
    LD (0, 5);
    LDF compare_lbl;
    AP 6;
    LDC 0;
    CGTE;
    SUB;
    RTN;
    (* le *)
    LABEL le_lbl;
    LDC 1;
    LDC 0;
    LDC 0;
    LDC 0;
    LDC 0;
    LD (0, 4);
    LD (0, 5);
    LDF compare_lbl;
    AP 6;
    LDC 0;
    CGT;
    SUB;
    RTN;
    (* gt *)
    LABEL gt_lbl;
    LDC 0;
    LDC 0;
    LDC 0;
    LDC 0;
    LD (0, 4);
    LD (0, 5);
    LDF compare_lbl;
    AP 6;
    LDC 0;
    CGT;
    RTN;
    (* ge *)
    LABEL ge_lbl;
    LDC 0;
    LDC 0;
    LDC 0;
    LDC 0;
    LD (0, 4);
    LD (0, 5);
    LDF compare_lbl;
    AP 6;
    LDC 0;
    CGTE;
    RTN;
  ]

let y_lbl = L "yYyYy"

let y = [
  LABEL y_lbl ;
  LD (0, 1) ;
  RTN
]

let rec interval n m =
  if n > m then [] else n :: interval (succ n) m


let lbl_y, generation_Y =
  let maximum_Y = ref 0 in
  (fun i ->
     maximum_Y := max !maximum_Y i ;
     L (Printf.sprintf "yYyYy_%03X" i)),
  (fun () ->
     List.map
       (fun i ->
          [LABEL (L (Printf.sprintf "yYyYy_%03X" i));
           LD (0, i-1);
           RTN])
       (interval 1 !maximum_Y)
     |> List.flatten)

let lbl_z, generation_Z =
  let maximum_Z = ref 0 in
  (fun i ->
     maximum_Z := max !maximum_Z i ;
     L (Printf.sprintf "zZzZz_%03X" i)),
  (fun () ->
     List.map
       (fun i ->
          LABEL (L (Printf.sprintf "zZzZz_%03X" i)) ::
          List.map (fun i -> LD (0, i)) (interval 0 (pred i)) @
          [RTN])
       (interval 1 !maximum_Z)
     |> List.flatten)

let lbl_restart, generate_restarters =
  let maximum_restarter = ref 0 in
  (fun i ->
     maximum_restarter := max !maximum_restarter i ;
     L (Printf.sprintf "rEsTaRt_%03X" i)),
  (fun () ->
     List.map
       (fun i ->
          (LABEL (L (Printf.sprintf "rEsTaRt_%03X" i))
           :: List.map
             (fun j -> LD (1, j))
             (interval 0 (pred i))
           @ [ LD (0, 0) ;
               LDC i ;
               ADD ;
               LD (1, i) ;
               TAP 1 ]))
       (interval 1 !maximum_restarter)
     |> List.flatten)

let lbl_build_res = L "build_res"
let lbl_call_res = L "call_res"
let build_res =
  [ LABEL lbl_build_res;
    LD (0, 4);
    CAR;
    LD (0, 4);
    CDR;
    ST (0, 0);
    LDF lbl_call_res;
    CONS;
    RTN;
    LABEL lbl_call_res;
    LD (0, 1);
    LD (0, 0);
    LDC 2;
    LD (1, 0);
    TAP 1;
 ]

let builtins () =
  block 255
  @ generation_Y ()
  @ generation_Z ()
  @ generate_restarters ()
  @ build_res
  @ caml_compare
  @ caml_make_vect
