open GccTypes

(* The main RAM was—for its time—impressively large, and was able to
   hold 10 million CONS cells to be used by the Lambda-Man AI. The three
   stacks and the heap shared the RAM used. The data stack used one CONS
   cell for every two data values. The control stack used one CONS cell
   per entry. Each environment frame used one CONS cell plus half of the
   frame size (rounded down). Again, trying to allocate memory over this
   limit resulted in an error, which probably lead to catastrophic
   failure.
*)

(* let debug = ref (fun _ -> ()) *)
(* let set_debug d = debug := d *)

type 'a info =
  | Val of 'a GccTypes.value
  | Error of string

type 'a debugger =
    'a runner -> 'a info -> 'a GccTypes.state ->
    string list -> 'a list -> unit

and 'a tracer =
    'a runner -> 'a GccTypes.state -> unit

and 'a runner =
    { program : 'a program;
      cast : 'a -> int;
      back : int -> 'a;
      debug : 'a debugger;
      trace : 'a tracer;
      pack_exn : 'a state -> any_state;
      unpack : any_state -> 'a state;
    }

let fault s = raise (FAULT s)

let gcn n s = s.used <- s.used + n
let gc s = gcn 1 s

let push_data s v = gc s; { s with data = v :: s.data  }
let pop_data s =
  match s.data with
  | [] -> fault (EMPTY_STACK "data")
  | v :: data -> v, { s with data }

let push_control s v = gc s; { s with control = v :: s.control  }
let pop_control runner s =
  match s.control with
  | [] -> fault (EMPTY_STACK "control")
  | v :: control -> v, { s with control }

type runner_state =
  | Pack : 'a runner -> runner_state

let rec run_ins_gen_loop :
type t. t runner -> t state -> string list -> t list -> t state =
    fun (runner: t runner) (s:t state) comments labels ->
  let addr = runner.cast in
  let back = runner.back in
  let pack = runner.pack_exn in
  assert(s.program == runner.program);

  runner.trace runner s;

  let ins = s.program.(s.pc) in
  let s = { s with pc = s.pc + 1 } in

  let check_closure x = match x with
    | TAG_CLOSURE (f, e) -> (f,e)
    | _ ->
       runner.debug runner (Val x) s [] [];
       fault (TAG_MISMATCH "closure") in

  let check_int x = match x with
    | TAG_INT n -> n
    | _ ->
       runner.debug runner (Val x) s [] [];
       fault (TAG_MISMATCH "int") in
  let check_cons x = match x with
    | TAG_CONS (x,y) -> (x,y)
    | _ ->
       runner.debug runner (Val x) s [] [];
       fault (TAG_MISMATCH "cons") in

  let check_join = function
    | TAG_JOIN n -> n
    | _ ->
       fault (TAG_MISMATCH "join") in

  let bin_int op s =
    let y, s = pop_data s in
    let x, s = pop_data s in
    let x = check_int x in
    let y = check_int y in
    let z = op x y in
    push_data s (TAG_INT z) in

  match ins with
  | LDC n -> push_data s (TAG_INT n)
  | LD (n,i) ->
    let rec iter e n =
      if n > 0 then
        match e.parent with
        | None -> fault (EMPTY_FRAME "LD")
        | Some e ->
           iter e (n-1)
      else
        (if i < 0 || i > Array.length e.values
         then
           (let rec frames { values; parent } =
              let r = match parent with
                | None -> []
                | Some f -> frames f
              in
              Array.length values :: r
            in
            let f = String.concat " :: " (List.map (string_of_int) (frames s.env)) in
            runner.debug
              runner
              (Error (Printf.sprintf
                        "index out of bound: LD %i %i (size %i, %s)"
                        n i (Array.length e.values) f))
              s comments labels);
         push_data s e.values.(i))
    in
    iter s.env n
  | ADD -> bin_int (+) s
  | SUB -> bin_int (-) s
  | MUL -> bin_int ( * ) s
  | DIV -> bin_int (/) s
  | CEQ -> bin_int (fun x y -> if x = y then 1 else 0) s
  | CGT -> bin_int (fun x y -> if x > y then 1 else 0) s
  | CGTE -> bin_int (fun x y -> if x >= y then 1 else 0) s
  | ATOM ->
    let x, s = pop_data s in
    let z = match x with
      | TAG_INT _ -> 1
      | _ -> 0 in
    push_data s (TAG_INT z)
  | CONS ->
    let y, s = pop_data s in
    let x, s = pop_data s in
    gc s;
    let z = TAG_CONS (x, y) in
    push_data s z

  | CAR ->
    let x, s = pop_data s in
    let (y, _ ) = check_cons x in
    push_data s y
  | CDR ->
    let x, s = pop_data s in
    let (_, y ) = check_cons x in
    push_data s y

  | SEL (t, f) ->
    let x, s = pop_data s in
    let x = check_int x in
    let s = push_control s (TAG_JOIN (back s.pc)) in
    let pc = if x = 0 then addr f else addr t in
    { s with pc }

  | JOIN ->
    let x, s = pop_control runner s in
    let pc = addr (check_join x) in
    { s with pc }

  | LDF f ->
    gc s;
    push_data s (TAG_CLOSURE (f, s.env))

  | AP n ->
    let x, s = pop_data s in
    let f, e = check_closure x in
    let rec iter s n values =
      if n = 0 then
        Array.of_list values, s
      else
        let x, s = pop_data s in
        iter s (n-1) (x :: values)
    in
    let values, s = iter s n [] in
    gcn n s;
    let fp = {
      parent = Some e;
      values = values;
    } in
    let s = push_control s (TAG_RET (back s.pc, s.env)) in
    { s with env = fp; pc = addr f }

  | RTN ->
    let x, s = pop_control runner s in
    begin
      match x with
      | TAG_STOP -> raise (MACHINE_STOP (pack s))
      | TAG_JOIN _ -> fault (TAG_MISMATCH "ret")
      | TAG_RET (x, frame) ->
         { s with pc = addr x; env = frame }
    end

  | DUM n ->
    gcn n s;
    let fp = { parent = Some s.env; values = Array.create n TAG_DUM; } in
    { s with env = fp }

  | RAP n ->
    let x, s = pop_data s in
    let (f, fp) = check_closure x in
    if fp.values <> [||] &&
       fp.values.(0) <> TAG_DUM then fault (TAG_MISMATCH "dum");
    if n <> Array.length fp.values then
      fault (FRAME_MISMATCH (n, Array.length fp.values));
    let rec iter s n values =
      if n > 0 then
        let x, s = pop_data s in
        values.(n-1) <- x;
        iter s (n-1) values
      else s
    in
    let s = iter s n fp.values in
    begin
      match fp.parent with
      | None -> fault (EMPTY_FRAME "RAP")
      | Some fpp ->
         let s = push_control s (TAG_RET (back s.pc, fpp)) in
         { s with env = fp; pc = addr f }
    end

  | STOP ->
    raise (MACHINE_STOP (pack s))

  | TSEL (t,f) ->
    let x, s = pop_data s in
    let x = check_int x in
    let pc = if x = 0 then addr f else addr t in
    { s with pc }

  | TAP n ->
    let x, s = pop_data s in
    let (f, e) = check_closure x in
    let rec iter s n values =
      if n = 0
      then Array.of_list values, s
      else
        let x, s = pop_data s in
        iter s (n-1) (x :: values)
    in
    let values, s = iter s n [] in
    gcn n s;
    let fp = { parent = Some e; values; } in
    { s with pc = addr f; env = fp }

  | TRAP n ->
    let x, s = pop_data s in
    let f, fp = check_closure x in
    if fp.values <> [||] &&
       fp.values.(0) <> TAG_DUM then fault (TAG_MISMATCH "dum");
    if n <> Array.length fp.values then
      fault (FRAME_MISMATCH (n, Array.length fp.values));
    let rec iter s n values =
      if n > 0 then
        let x, s = pop_data s in
        values.(n-1) <- x;
        iter s (n-1) values
      else s
    in
    let s = iter s n fp.values in
    let _fpp = fp.parent in
    { s with env = fp; pc = addr f }

  | ST (n,i) ->
    let fp = s.env in
    let rec iter fp n =
      if n = 0
      then fp
      else
        match fp.parent with
        | None -> fault (EMPTY_FRAME "ST")
        | Some e -> iter e (n-1)
    in
    let fp = iter fp n in
    let v, s = pop_data s in
    (if i < 0 || i > Array.length fp.values
     then
       (let rec frames { values; parent } =
         let r = match parent with
           | None -> []
           | Some f -> frames f
         in
         Array.length values :: r
        in
        let f = String.concat " :: " (List.map (string_of_int) (frames s.env)) in
        runner.debug
          runner
          (Error (Printf.sprintf
                    "index out of bound: LD %i %i (size %i, %s)"
                    n i (Array.length fp.values) f))
          s comments labels));
    fp.values.(i) <- v;
    s

  | DBUG ->
    let x, s = pop_data s in
    runner.debug runner (Val x) s comments labels;
    s
  | BRK ->
    raise (BREAKPOINT (pack s))

  | COMMENT c ->
     run_ins_gen_loop runner s (c::comments) labels
  | LABEL l ->
     run_ins_gen_loop runner s comments (l::labels)

let run_ins_gen runner state =
  run_ins_gen_loop runner state [] []

let empty_frame () =
  { parent = None; values = [||]; }

let init_state program =
  { data = [];
    control = [TAG_STOP];
    env = empty_frame ();
    pc = 0;
    program = program;
    used = 0;
 }

type 'a closure = ('a * 'a GccTypes.frame)

type 'a run_state =
  | Run of 'a GccTypes.state
  | Stop of GccTypes.any_state

let get_result (state:'a GccTypes.state) =
  Printf.eprintf "Approx space used %d\n%!" state.used;
  match state.data with
  | [] -> None
  | h :: t -> Some h

let int_runner program =
  { program;
    cast = (fun x -> x);
    back = (fun x -> x);
    debug = (fun _ _ _ _ _ -> ());
    trace = (fun _ _ -> ());
    pack_exn = (fun x -> Int_state x);
    unpack = (function Int_state s -> s | _ -> assert false) }

let pack_int state =
  Pack (int_runner state)

module IntMap = Map.Make(struct type t = int let compare (x:int) y = compare x y end)

let arg_runner ?(debug=((fun _ _ _ _ _ -> ()): arg debugger)) ?(trace=((fun _ _ -> ()):arg tracer)) program =
  let map = GccPrint.make_label_map program in
  let back_map = StringMap.fold (fun s i acc -> IntMap.add i s acc) map IntMap.empty in
  let find = function
    | I i -> i
    | L s ->
        try StringMap.find s map
        with Not_found ->
          Printf.eprintf "Missing label: %s\n%!" s; exit 2
  in
  let back_find i =
    try L (IntMap.find i back_map) with
    | Not_found -> I i
  in
  { program;
    cast = find;
    back = back_find;
    debug;
    trace;
    pack_exn = (fun x -> Arg_state x);
    unpack = (function Arg_state s -> s | _ -> assert false);}

let pack_label state =
  Pack (arg_runner state)

let run_ins (state:int state) =
  let runner = int_runner state.program in
  let state = run_ins_gen runner state in
  state

type 'a r =
  | R of 'a state
  | S of any_state

let count = ref 0

let rec instrs_runner : type t. t runner -> t state -> t state = fun r s ->
  incr count;
  match
    try R (run_ins_gen r s)
    with GccTypes.MACHINE_STOP state -> S state
  with
  | R state -> instrs_runner r state
  | S state -> r.unpack state

let get_count () =
  let n = !count in
  count := 0;
  n

let run_instrs program ((pointer, frame):'a closure)
               args : 'a GccTypes.value option =
  let state = init_state program in
  let frame = { parent = Some frame; values = args; } in
  let state =
    { state with
      control = [TAG_STOP];
      pc = pointer;
      env = frame } in
  let pack = int_runner program in
  let state = instrs_runner pack state in
  get_result state

let run_instrs' program ((pointer, frame):'a closure)
               args : 'a GccTypes.value option =
  let state = init_state program.program in
  let frame = { parent = Some frame; values = args; } in
  let state =
    { state with
      control = [TAG_STOP];
      pc = program.cast pointer;
      env = frame } in
  let state = instrs_runner program state in
  get_result state

let init_state' r = init_state r.program

let entry_point r = r.back 0


let position_name program pc =
  let open GccTypes in
  let rec aux pc =
    if pc = 0
    then "program start", 0
    else
      match program.(pc) with
      | LABEL (I i) -> Printf.sprintf "pos %i" i, 0
      | LABEL (L s) -> s, 0
      | _ ->
         let s, i = aux (pc-1) in
         s, (i+1)
  in
  let s, i = aux pc in
  if i = 0
  then s
  else Printf.sprintf "%s + %i" s i

let print_value runner ppf v =
  let rec aux i ppf v =
    if i = 0
    then Format.fprintf ppf "..."
    else
      match v with
      | TAG_INT i -> Format.fprintf ppf "Int %i" i
      | TAG_CONS (v1,v2) ->
         Format.fprintf ppf "Cons(@[@ %a,@ %a@])"
                        (aux (i-1)) v1 (aux (i-1)) v2
      | TAG_DUM -> Format.fprintf ppf "Dummy"
      | TAG_CLOSURE (pc, frame) ->
         let pc = runner.cast pc in
         let name = position_name runner.program pc in
         Format.fprintf ppf "Closure( %s )" name
  in
  let aux_for_map ppf l =
    let rec loop1 ppf = function
      | TAG_INT 0 -> ()
      | TAG_CONS (hd, tl) ->
          Format.fprintf ppf "%a\n%a" loop2 hd loop1 tl
      | _ -> Format.fprintf ppf "... WTF..."
    and loop2 ppf = function
      | TAG_INT 0 -> ()
      | TAG_CONS (TAG_INT 1, tl) ->
          Format.fprintf ppf "#%a" loop2 tl
      | TAG_CONS (TAG_INT 0, tl) ->
          Format.fprintf ppf " %a" loop2 tl
      | _ -> Format.fprintf ppf "... WTF..."
    in loop1 ppf l
  in
  let aux_for_array ppf l =
    let rec loop1 ppf = function
      | TAG_INT 0 -> ()
      | TAG_CONS (hd, tl) ->
          Format.fprintf ppf "%a\n%a" loop2 hd loop1 tl
      | _ -> Format.fprintf ppf "... WTF..."
    and loop2 ppf = function
      | TAG_INT 0 -> ()
      | TAG_CONS (TAG_INT 0, tl) ->
          Format.fprintf ppf "    %a" loop2 tl
      | TAG_CONS (TAG_INT n, tl) ->
          Format.fprintf ppf "%3d %a" n loop2 tl
      | _ -> Format.fprintf ppf "... WTF..."
    in loop1 ppf l
  in
  match v with
  | TAG_CONS (TAG_INT 1665_1665,v) -> aux_for_array ppf v
  | TAG_CONS (TAG_INT 1664_1664,v) -> aux_for_map ppf v
  | _ -> aux 4 ppf v

let simple_debug : Format.formatter -> GccTypes.arg debugger
  = fun ppf runner value (state:arg state) comments labels ->
  let position = position_name state.program state.pc in
  let comments =
    match comments with
    | [] -> ""
    | _ -> " " ^ String.concat ", " comments in
  let print_info ppf = function
    | Val v -> print_value runner ppf v
    | Error s -> Format.fprintf ppf "error: %s" s
  in
  let txtins = (GccPrint.show_ins state.program.(state.pc)) in
  Format.fprintf ppf
    "%s (%s)%s %a@."
    txtins
    position comments
    print_info value

let simple_trace : Format.formatter -> GccTypes.arg tracer
  = fun ppf runner (state:arg state) ->
  let position = position_name state.program state.pc in
  Format.fprintf ppf
    "%20s: %s@."
    position
    (GccPrint.show_ins state.program.(state.pc))
