open Format

type instr =
  | Jmp of rvalue
  | Set of lvalue * rvalue
  | Arith of arith_op * lvalue * rvalue
  | Cmp of cmp_op * pos * rvalue * rvalue
  | Interrupt of int
  | Halt
  | Label of string
and pos =
  | Addr of int
  | Symbol of string
and rvalue =
  | Pos of pos
  | Reg of reg
  | Cst of int
  | Mem_at_reg of reg
  | Mem_at_cst of int
  | Pc
and lvalue =
  | Set_reg of reg
  | Set_mem_at_reg of reg
  | Set_mem_at_cst of int
and arith_op =
  | Add | Sub | Mul | Div
  | And | Or | Xor
and cmp_op =
  | Lt | Eq  | Gt
and reg = int
and interrupt = int

let code = ref []
let text =
  let fp = open_in Sys.argv.(1) in
  let size = in_channel_length fp in
  let text = String.make (size + 1) '\000' in
  really_input fp text 0 size ;
  close_in fp ;
  text

let ppf = std_formatter

let rec spaces acc p i =
  match String.unsafe_get text i with
  | ' ' | '\t' -> spaces acc (i + 1) (i + 1)
  | 'a'..'z' | 'A'..'Z' -> ident acc i i
  | '@' -> label acc i (i + 1)
  | '0'..'9' -> number acc i i
  | '[' -> spaces (`O :: acc) (i + 1) (i + 1)
  | ']' -> (* very nice design *)
      (match acc with
       | `W r :: `O :: tl -> spaces (`W' r :: tl) (i + 1) (i + 1)
       | `N n :: `O :: tl -> spaces (`N' n :: tl) (i + 1) (i + 1)
       | _ -> failwith "syntax error, unexpected ']'")
  | '=' -> spaces (`E :: acc) (i + 1) (i + 1)
  | ',' -> spaces (`C :: acc) (i + 1) (i + 1)
  | '\n' -> spaces (`R :: acc) (i + 1) (i + 1)
  | ';' -> comment acc i i
  | '\000' -> List.rev (`R :: acc)
  | c -> failwith (Printf.sprintf "unexpected character %C" c)
and word p i = `W (String.uppercase (String.sub text p (i - p)))
and lab p i = `L (String.uppercase (String.sub text p (i - p)))
and num p i = `N (int_of_string (String.sub text p (i - p)))
and ident acc p i =
  match String.unsafe_get text i with
  | 'a'..'z' | 'A'..'Z' -> ident acc p (i + 1)
  | _ -> spaces (word p i :: acc) i i
and label acc p i =
  match String.unsafe_get text i with
  | 'a'..'z' | 'A'..'Z' -> label acc p (i + 1)
  | _ -> spaces (lab p i :: acc) i i
and number acc p i =
  match String.unsafe_get text i with
  | '0'..'9' -> number acc p (i + 1)
  | _ -> spaces (num p i :: acc) i i
and comment acc p i =
  match String.unsafe_get text i with
  | '\n' | '\000' -> spaces acc i i
  | c -> comment acc p (i + 1)

let tokens = spaces [] 0 0

let rec parse toks =
  let regn n =
    if String.length n > 1 then
      failwith (Printf.sprintf "bad register name %S" n)
    else
    let r = Char.code n.[0] - Char.code 'A' in
    if r < 0 || r > 7 then
      failwith (Printf.sprintf "bad register name %S" n)
    else r
  in
  let arith_op = function
    | "ADD" -> Add | "SUB" -> Sub | "MUL" -> Mul | "DIV" -> Div
    | "AND" -> And | "OR" -> Or | "XOR" -> Xor
    | _ -> assert false
  and cmp_op = function
    | "JLT" -> Lt | "JEQ" -> Eq | "JGT" -> Gt
    | _ -> assert false
  and rvalue = function
    | `W "PC" -> Pc
    | `W r -> Reg (regn r)
    | `N n -> Cst n
    | `W' r -> Mem_at_reg (regn r)
    | `N' n -> Mem_at_cst n
    | `L l -> Pos (Symbol l)
  and lvalue = function
    | `W r -> Set_reg (regn r)
    | `W' r -> Set_mem_at_reg (regn r)
    | `N' n -> Set_mem_at_cst n
  in
  match toks with
  | `W "NOP" :: `R :: tl ->
      code := Set (Set_reg 0, Reg 0) :: !code ;
      parse tl
  | `W "JMP" :: `N cst :: `R :: tl
  | `W "MOV" :: `W "PC" :: `C :: `N cst :: `R :: tl ->
      code := Jmp (Pos (Addr cst)) :: !code ;
      parse tl
  | `W "JMP" :: `L l :: `R :: tl
  | `W "MOV" :: `W "PC" :: `C :: `L l :: `R :: tl ->
      code := Jmp (Pos (Symbol l)) :: !code ;
      parse tl
  | `W "JMP" :: `W "PC" :: `R :: tl
  | `W "MOV" :: `W "PC" :: `C :: `W "PC" :: `R :: tl ->
      code := Jmp Pc :: !code ;
      parse tl
  | `W "JMP" :: `W reg :: `R :: tl
  | `W "MOV" :: `W "PC" :: `C :: `W reg :: `R :: tl ->
      code := Jmp (Reg (regn reg)) :: !code ;
      parse tl
  | (`W _ | `W' _ | `N' _ as dst) :: `E :: (`W' _ | `N' _ | `W _ | `N _ | `L _ as src) :: `R :: tl
  | `W "MOV"
    :: (`W _ | `W' _ | `N' _ as dst) :: `C
    :: (`W' _ | `N' _ | `W _ | `N _ | `L _ as src) :: `R :: tl ->
      code := Set (lvalue dst, rvalue src) :: !code ;
      parse tl
  | `W "INC" :: (`W _ | `W' _ | `N' _ as dst) :: `R :: tl ->
      code := Arith (Add, lvalue dst, Cst 1) :: !code ;
      parse tl
  | `W "DEC" :: (`W _ | `W' _ | `N' _ as dst) :: `R :: tl ->
      code := Arith (Sub, lvalue dst, Cst 1) :: !code ;
      parse tl
  | `W ("ADD" | "MUL" | "DIV" | "SUB" | "AND" | "OR" | "XOR" as op)
    :: (`W _ | `W' _ | `N' _ as dst) :: `C
    :: (`W' _ | `N' _ | `W _ | `N _ | `L _  as src) :: `R :: tl ->
      code := Arith (arith_op op, lvalue dst, rvalue src) :: !code ;
      parse tl
  | `W ("JLT" | "JEQ" | "JGT" as op)
    :: `N addr :: `C
    :: (`W' _ | `N' _ | `W _ | `N _ | `L _  as x) :: `C
    :: (`W' _ | `N' _ | `W _ | `N _ | `L _  as y) :: `R :: tl ->
      code := Cmp (cmp_op op, Addr addr, rvalue x, rvalue y) :: !code ;
      parse tl
  | `W ("JLT" | "JEQ" | "JGT" as op)
    :: `L addr :: `C
    :: (`W' _ | `N' _ | `W _ | `N _ | `L _  as x) :: `C
    :: (`W' _ | `N' _ | `W _ | `N _ | `L _  as y) :: `R :: tl ->
      code := Cmp (cmp_op op, Symbol addr, rvalue x, rvalue y) :: !code ;
      parse tl
  | `W "INT" :: `N n :: `R :: tl ->
      code := Interrupt n :: !code ;
      parse tl
  | `W "HLT" :: `R :: tl ->
      code := Halt :: !code ;
      parse tl
  | `R :: tl ->
      parse tl
  | `L l :: tl ->
      code := Label l :: !code ;
      parse tl
  | `E :: _ -> failwith (Printf.sprintf "syntax error, unexpected '='")
  | `O :: _ -> failwith (Printf.sprintf "syntax error, unexpected '['")
  | `C :: _ -> failwith (Printf.sprintf "syntax error, unexpected coma")
  | `N n :: _ -> failwith (Printf.sprintf "syntax error, unexpected %d" n)
  | `W n :: _ -> failwith (Printf.sprintf "syntax error, unexpected %s" n)
  | `N' n :: _ -> failwith (Printf.sprintf "syntax error, unexpected [%d]" n)
  | `W' n :: _ -> failwith (Printf.sprintf "syntax error, unexpected [%s]" n)
  | [] -> ()

let _ = parse tokens

let code = List.rev !code

let labels = Hashtbl.create 256

let _ =
  let i = ref 0 in
  List.iter
    (function
       | Label s ->
           if Hashtbl.mem labels s then
             failwith ("DUPLICATE LABEL " ^ String.uppercase s) ;
           Hashtbl.replace labels s !i
       | _ -> incr i)
    code

let rec print_instr ppf = function
  | Jmp rv -> fprintf ppf "MOV pc, %a" print_rvalue rv
  | Set (lv, rv) -> fprintf ppf "MOV %a, %a" print_lvalue lv print_rvalue rv
  | Arith (op, lv, rv) -> fprintf ppf "%a %a, %a" print_arith_op op print_lvalue lv print_rvalue rv
  | Cmp (op, pos, rv1, rv2) -> fprintf ppf "%a %a, %a, %a" print_cmp_op op print_pos pos print_rvalue rv1 print_rvalue rv2
  | Interrupt n -> fprintf ppf "INT %i" n
  | Halt -> fprintf ppf "HLT"
  | Label _ -> assert false
and print_pos ppf = function
  | Addr i -> fprintf ppf  "%i" i
  | Symbol n ->
      try fprintf ppf  "%i" (Hashtbl.find labels n)
      with Not_found -> failwith ("UNBOUND LABEL " ^ String.uppercase n)
and print_rvalue ppf = function
  | Pos pos -> print_pos ppf pos
  | Reg reg -> print_reg ppf reg
  | Cst i -> fprintf ppf  "%i" i
  | Mem_at_reg reg -> fprintf ppf "[%a]" print_reg reg
  | Mem_at_cst i -> fprintf ppf "[%i]" i
  | Pc -> fprintf ppf "PC"
and print_lvalue ppf = function
  | Set_reg  reg -> print_reg ppf reg
  | Set_mem_at_reg reg -> fprintf ppf "[%a]" print_reg reg
  | Set_mem_at_cst i -> fprintf ppf "[%i]" i
and print_arith_op ppf = function
  | Add -> fprintf ppf "ADD"
  | Sub -> fprintf ppf "SUB"
  | Mul -> fprintf ppf "MUL"
  | Div -> fprintf ppf "DIV"
  | And -> fprintf ppf "AND"
  | Or -> fprintf ppf "OR"
  | Xor -> fprintf ppf "XOR"
and print_cmp_op ppf = function
  | Lt -> fprintf ppf "JLT"
  | Eq -> fprintf ppf "JEQ"
  | Gt -> fprintf ppf "JGT"
and print_reg ppf i =
  fprintf ppf "%c" "ABCDEFGH".[i]

let _ =
  let i = ref 0 in
  List.iter
    (function
      | Label _ -> ()
      | instr -> fprintf ppf "%a\t; %i\n" print_instr instr !i ; incr i)
    code

